from typing import Optional, Any

from pypst.utils import render_mapping, render_sequence


class Cell:
    """
    A cell element that can be rendered as a table cell.

    Normally, the user does not need to create a cell directly.
    When creating a table from a data frame, the table will automatically create cells.
    These cells can then be modified to style them individually.

    Args:
        value: The content of the cell.
        rowspan: The number of rows the cell spans.
        colspan: The number of columns the cell spans.
        fill: The background color of the cell.
        align: The alignment of the cell content.
        stroke: The stroke style of the cell.

    Examples:
        >>> cell = Cell(
        ...     "Value",
        ...     rowspan=2,
        ...     colspan=3,
        ...     fill="red",
        ...     align="center",
        ...     stroke="black"
        ... )
        >>> print(cell)
        [#table.cell(rowspan: 2, colspan: 3, fill: red, align: center, stroke: black)[Value]]
    """

    def __init__(
        self,
        value: Optional[Any] = None,
        rowspan: int = 1,
        colspan: int = 1,
        fill: Optional[str] = None,
        align: Optional[str] = None,
        stroke: Optional[str | list | dict] = None,
    ) -> None:
        self.value = value
        self.rowspan = rowspan
        self.colspan = colspan
        self.fill = fill
        self.align = align
        self.stroke = stroke

    def __str__(self) -> str:
        return self.render()

    def __repr__(self) -> str:
        return (
            f"Cell(value={self.value}, rowspan={self.rowspan}, colspan={self.colspan}, "
            f"fill={self.fill}, align={self.align}, stroke={self.stroke})"
        )

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Cell):
            return False

        return (
            self.value == other.value
            and self.rowspan == other.rowspan
            and self.colspan == other.colspan
            and self.fill == other.fill
            and self.align == other.align
            and self.stroke == other.stroke
        )

    def render(self) -> str:
        """
        Render the cell to string.

        The cell function is surrounded by a content block to fit the table syntax.

        Returns:
            The rendered cell.
        """
        args = []
        if self.rowspan > 1:
            args.append(f"rowspan: {self.rowspan}")
        if self.colspan > 1:
            args.append(f"colspan: {self.colspan}")
        if self.fill is not None:
            args.append(f"fill: {self.fill}")
        if self.align is not None:
            args.append(f"align: {self.align}")
        if self.stroke is not None:
            stroke = self.stroke
            if isinstance(self.stroke, dict):
                stroke = render_mapping(self.stroke)
            elif isinstance(self.stroke, list):
                stroke = render_sequence(self.stroke)
            args.append(f"stroke: {stroke}")

        cell = "[]" if self.value is None else f"[{self.value}]"
        if args:
            cell = f"[#table.cell({', '.join(args)}){cell}]"

        return cell



from dataclasses import dataclass, field

from pypst.document import Document
from pypst.renderable import Renderable
from pypst.utils import Function, dataclass_fields_to_render, render


@dataclass
class Heading(Function):
    """
    A Heading element.

    If only the body and level are provided,
    the heading will be rendered as in Markdown.
    If any other attribute is provided, the heading will be rendered as a function.

    Args:
        body: The content of the heading.
        level: The absolute nesting depth of the heading, starting from one.
        depth: The relative nesting depth of the heading, starting from one.
        offset: The starting offset of each heading's level,
                used to turn its relative depth into its absolute level.
        numbering: The numbering scheme, for example, `"1.1"`.
        supplement: The supplement for the heading.
        outlined: Whether the heading appears in the outline.
        bookmarked: Whether the heading appears as a bookmark in PDF.

    Examples:
        >>> h = Heading("Heading 1", level=1)
        >>> print(h.render())
        = Heading 1

        >>> h = Heading("Heading 1", level=3)
        >>> print(h.render())
        === Heading 1

        >>> h = Heading('"Heading 1"', depth=2, offset=1)
        >>> print(h.render())
        #heading("Heading 1", depth: 2, offset: 1)
    """

    __is_function__ = True

    body: Renderable | str = field(metadata={"positional": True})
    level: int | None = None
    depth: int | None = None
    offset: int | None = None
    numbering: str | None = None
    supplement: Renderable | str | None = None
    outlined: bool | None = None
    bookmarked: bool | None = None

    def __post_init__(self) -> None:
        if self.level is not None and self.level < 1:
            raise ValueError("Level must be greater than 0")
        if self.depth is not None and self.depth < 1:
            raise ValueError("Depth must be greater than 0")
        if self.offset is not None and self.offset < 0:
            raise ValueError("Offset must be greater than or equal to 0")
        if (
            self.depth is not None or self.offset is not None
        ) and self.level is not None:
            raise ValueError("Level cannot be set if depth or offset is set")
        if not isinstance(self.body, (Renderable, str)):
            raise ValueError(f"Invalid body type: {type(self.body)}")
        if isinstance(self.body, Document):
            raise ValueError(
                "Document cannot be set as body, because it needs to be top-level"
            )

    def render(self) -> str:
        """
        Render the heading to a string.

        If only body and level are provided,
        the heading will be rendered as in Markdown.
        If any other attribute is provided, the heading will be rendered as a function.

        Returns:
            The rendered heading element.
        """
        if self.level is not None and len(list(dataclass_fields_to_render(self))) == 2:
            # remove unnecessary quotes, because Markdown style is not in code mode
            body = render(self.body).strip('"')
            return f"{'=' * self.level} {body}"
        else:
            return super().render()



from collections.abc import Iterable
from dataclasses import dataclass

from pypst.renderable import Renderable
from pypst.utils import render_code, render_fenced


@dataclass
class Functional:
    """
    Functional block wrapped in {} in Typst.

    Examples:
        >>> from pypst import SetRule, Content
        >>> content = Functional([SetRule("text", {"fill": "red"}), Content("This text will be red.")])
        >>> print(content.render())
        #{
          set text(fill: red)
          [This text will be red.]
        }

        >>> from pypst import SetRule, Content
        >>> content = Functional([SetRule("text", {"fill": "red"}), Content("This is page #here().page()")], context=True)
        >>> print(content.render())
        #context {
          set text(fill: red)
          [This is page #here().page()]
        }
    """

    body: str | Renderable | Iterable[str | Renderable] | None = None
    context: bool = False
    joint: str = "\n"
    indent: int = 2

    def __post_init__(self) -> None:
        if isinstance(self.body, Functional):
            self.context = self.context or self.body.context
            self.body = self.body.body

    def __str__(self) -> str:
        return self.render()

    def render(self) -> str:
        """
        Render the functional block to a string.
        """
        return render_fenced(
            body=self.body,
            context=self.context,
            joint=self.joint,
            indent=self.indent,
            start="{",
            end="}",
            render_fn=render_code,
        )



from dataclasses import dataclass, field
from typing import Literal

from pypst.utils import Function


@dataclass
class Image(Function):
    """
    Image object to add image elements.

    The image file needs to be in the provided path.
    The path is interpreted as relative to the Typst file.

    Args:
        path: Path to the image file.
        format: The image format.
        width: The width the image should occupy.
        height: The height the image should occupy.
        alt: The image alt text.
        fit: How the image fills the space defined by `width` and `height`.

    Examples:
        >>> image = Image(path="image.png", width="100%", height="50%")
        >>> print(image.render())
        #image("image.png", width: 100%, height: 50%)
    """

    __is_function__ = True

    path: str = field(metadata={"positional": True})
    format: Literal["png", "jpg", "gif", "svg"] | str | None = None
    width: str | None = None
    height: str | None = None
    alt: str | None = None
    fit: Literal["cover", "contain", "stretch"] | str | None = None

    def __post_init__(self) -> None:
        if not self.path.startswith('"'):
            self.path = f'"{self.path}"'

        if self.format is not None and not self.format.startswith('"'):
            self.format = f'"{self.format}"'

        if self.fit is not None and not self.fit.startswith('"'):
            self.fit = f'"{self.fit}"'



from dataclasses import dataclass, field
from typing import Optional

from pypst import utils
from pypst.renderable import Plain, Renderable


class Document:
    """
    Represents a Typst document serving as the top-level container.

    A document contains other renderable objects and necessary imports.
    It is only necessary if multiple renderable objects need to be combined
    or third-party files need to be imported.

    Examples:
            >>> doc = Document("Hello, world!")
            >>> print(doc.render())
            Hello, world!

            >>> doc = Document(["Hello,", "world!"])
            >>> print(doc.render())
            Hello,
            <BLANKLINE>
            world!

            >>> doc = Document()
            >>> doc.add("Hello,")
            >>> doc.add("world!")
            >>> print(doc.render())
            Hello,
            <BLANKLINE>
            world!

            >>> doc = Document()
            >>> doc.add_import("module")
            >>> doc.add("Hello, world!")
            >>> print(doc.render())
            #import "module"
            <BLANKLINE>
            Hello, world!
    """

    _body: list[Renderable | str]
    imports: list["Import"]

    def __init__(
        self, body: Optional[Renderable | str | list[Renderable | str]] = None
    ) -> None:
        """
        Create a new document with an optional body.

        The body can be a single renderable object, a string,
        or a list of renderable objects and strings.

        Args:
            body: The body of the document.
        """
        self._body = []
        self.imports = []

        if body is None:
            body = []
        elif not isinstance(body, list):
            body = [body]
        for b in body:
            self.add(b)

    @property
    def body(self) -> list[Renderable | str]:
        """The body of the document."""
        return self._body

    def add(self, element: Renderable | str) -> None:
        """
        Add a body element to the document.

        Any object of the `pypst` library with a `render` method can be added.
        Simple text paragraphs or unsupported Typst elements can be added as
        strings.
        Adding an `Import` object is equivalent to calling `add_import`.

        Args:
            element: The new body element.
        """
        if not isinstance(element, (Renderable, str)):
            raise ValueError(f"Invalid body type: {type(element)}")
        elif isinstance(element, Document):
            raise ValueError("Document cannot be set as value of another document")
        elif isinstance(element, Import):
            self.imports.append(element)
        else:
            self._body.append(element)

    def add_import(
        self, module: str | Plain, members: Optional[list[str]] = None
    ) -> None:
        """
        Add an import to the document.

        The module is wrapped in quotes upon rendering by default,
        but can be replaced by a `Plain` value to facilitate imports from a module value.
        If members are provided,
        only those members will be imported as `#import <module>: <list of members>`.
        If no members are provided, the entire module will be imported as
        `#import <module>`.
        Imports are rendered at the top of the document.

        Args:
            module: The module to import.
            members: The optional list of members to import.
        """
        self.imports.append(Import(module, members or []))

    def render(self) -> str:
        """
        Render the document.

        The rendered document has all imports at the top followed by the body.
        Each element in the body is rendered in the order it was added.
        A blank line separates each body element.

        Returns:
            The rendered document string.
        """
        imports = "\n".join([i.render() for i in self.imports])
        body = "\n\n".join(utils.render(b) for b in self.body)

        if imports:
            return f"{imports}\n\n{body}"
        else:
            return body


@dataclass
class Import:
    """
    Typst import statement to add to the document.

    The module is wrapped in quotes upon rendering by default,
    but can be replaced by a `Plain` value to facilitate imports from a module value.
    If members are provided,
    only those members will be imported as `#import <module>: <list of members>`.
    If no members are provided, the entire module will be imported as
    `#import <module>`.
    Imports are rendered at the top of the document.

    Args:
        module: The module to import.
        members: The optional list of members to import.
    """

    module: str | Plain
    members: list[str] = field(default_factory=list)

    def render(self) -> str:
        if isinstance(self.module, Plain):
            module = self.module.render()
        else:
            module = self.module.strip('"')
            module = f'"{module}"'
        rendered_import = f"#import {module}"
        if self.members:
            if "*" in self.members and len(self.members) > 1:
                raise ValueError(
                    f"Error while rendering import for '{self.module}'. "
                    "Cannot import all and specific members at the same time"
                )
            members = ", ".join(self.members)
            rendered_import += f": {members}"

        return rendered_import



import itertools
from dataclasses import dataclass
from typing import Optional

from pypst import utils
from pypst.document import Document
from pypst.renderable import Renderable


@dataclass
class Itemize:
    """
    An element that represents a bullet point list.

    If no arguments are provided, the list will be rendered as a Markdown list.
    Otherwise, the list will be rendered as a function.
    It is possible to nest lists in both Markdown and functional forms.

    Args:
        elements: List of elements to be rendered as bullet points.
        tight: Whether to use list or leading spacing between elements.
        marker: The bullet point markers to use.
        indent: The indent for each element.
        body_indent: The spacing between the marker and the element body.
        spacing: The spacing of a non-tight list.

    Examples:
        >>> itemize = Itemize(["First", "Second"])
        >>> print(itemize.render())
        - First
        - Second

        >>> itemize = Itemize(["First", "Second"], tight=True)
        >>> print(itemize.render())
        #list(
          tight: true,
          [First],
          [Second]
        )

        >>> itemize = Itemize(["First", Itemize(["Nested 1", "Nested 2"]), "Second"])
        >>> print(itemize.render())
        - First
          - Nested 1
          - Nested 2
        - Second

        >>> enumerate = Itemize(["First", Enumerate(["Nested 1", "Nested 2"]), "Second"])
        >>> print(enumerate.render())
        - First
          + Nested 1
          + Nested 2
        - Second

        >>> itemize = Itemize([])
        >>> itemize.add("First")
        >>> itemize.add("Second")
        >>> print(itemize.render())
        - First
        - Second
    """

    elements: list[Renderable | str]
    tight: Optional[bool] = None
    marker: Optional[str | list[str]] = None
    indent: Optional[str] = None
    body_indent: Optional[str] = None
    spacing: Optional[str] = None

    def __post_init__(self) -> None:
        _check_elements(self.elements)

    def add(self, element: Renderable | str) -> None:
        """
        Add an element to the bullet point list.

        Args:
            element: The new element.
        """
        _check_element(element)
        self.elements.append(element)

    def render(self) -> str:
        """
        Render the bullet point list to string.

        If any arguments are provided, the list will be rendered as a function.
        Otherwise, the list will be rendered as a Markdown list.

        Returns:
            The rendered list.
        """
        args = []
        if self.tight:
            args.append(f"tight: {utils.render(self.tight)}")
        if self.marker:
            args.append(f"marker: {utils.render(self.marker)}")
        if self.indent:
            args.append(f"indent: {self.indent}")
        if self.body_indent:
            args.append(f"body-indent: {self.body_indent}")
        if self.spacing:
            args.append(f"spacing: {self.spacing}")

        if args:
            rendered = _render_functional(args, self.elements, "list")
        else:
            rendered = _render_markdown(self.elements, "-")

        return rendered


@dataclass
class Enumerate:
    """
    An element that represents a numbered list.

    If no arguments are provided, the list will be rendered as a Markdown list.
    Otherwise, the list will be rendered as a function.
    It is possible to nest lists in both Markdown and functional forms.

    Args:
        elements: List of elements to be rendered as numbered points.
        tight: Whether to use list or leading spacing between elements.
        numbering: The numbering scheme to use.
        start: The starting number for the list.
        full: Whether to use full numbering with parent enumerations.
        indent: The indent for each element.
        body_indent: The spacing between the number and the element body.
        spacing: The spacing of a non-tight list.
        number_align: The alignment of the numbers.

    Examples:
        >>> enumerate = Enumerate(["First", "Second"])
        >>> print(enumerate.render())
        + First
        + Second

        >>> enumerate = Enumerate(["First", Enumerate(["Nested 1", "Nested 2"]), "Second"])
        >>> print(enumerate.render())
        + First
          + Nested 1
          + Nested 2
        + Second

        >>> enumerate = Enumerate(["First", Itemize(["Nested 1", "Nested 2"]), "Second"])
        >>> print(enumerate.render())
        + First
          - Nested 1
          - Nested 2
        + Second

        >>> enumerate = Enumerate([])
        >>> enumerate.add("First")
        >>> enumerate.add("Second")
        >>> print(enumerate.render())
        + First
        + Second
    """

    elements: list[Renderable | str]
    tight: Optional[bool] = None
    numbering: Optional[str] = None
    start: Optional[int] = None
    full: Optional[bool] = None
    indent: Optional[str] = None
    body_indent: Optional[str] = None
    spacing: Optional[str] = None
    number_align: Optional[str] = None

    def __post_init__(self) -> None:
        _check_elements(self.elements)

    def add(self, element: Renderable | str) -> None:
        """
        Add an element to the numbered list.

        Args:
            element: The new element.
        """
        _check_element(element)
        self.elements.append(element)

    def render(self) -> str:
        """
        Render the numbered list to string.

        If any arguments are provided, the list will be rendered as a function.
        Otherwise, the list will be rendered as a Markdown list.

        Returns:
            The rendered numbered list.
        """
        args = []
        if self.tight:
            args.append(f"tight: {utils.render(self.tight)}")
        if self.numbering:
            args.append(f"numbering: {self.numbering}")
        if self.start:
            args.append(f"start: {self.start}")
        if self.full:
            args.append(f"full: {utils.render(self.full)}")
        if self.indent:
            args.append(f"indent: {self.indent}")
        if self.body_indent:
            args.append(f"body-indent: {self.body_indent}")
        if self.spacing:
            args.append(f"spacing: {self.spacing}")
        if self.number_align:
            args.append(f"number-align: {self.number_align}")

        if args:
            rendered = _render_functional(args, self.elements, "enum")
        else:
            rendered = _render_markdown(self.elements, "+")

        return rendered


def _check_elements(elements: list[Renderable | str]) -> None:
    if isinstance(elements, list):
        if any(not isinstance(e, (Renderable, str)) for e in elements):
            raise ValueError("Encountered invalid element type in elements list")
        if any(isinstance(e, Document) for e in elements):
            raise ValueError("Document cannot be added as it needs to be top-level")
    else:
        raise ValueError(f"Invalid elements type {type(elements)}")


def _check_element(element: Renderable | str) -> None:
    if not isinstance(element, (Renderable, str)):
        raise ValueError(f"Invalid element type: {type(element)}")
    if isinstance(element, Document):
        raise ValueError("Document cannot be added as it needs to be top-level")


def _render_functional(
    args: list[str], elements: list[Renderable | str], function_name: str
) -> str:
    body = []
    for e in elements:
        b = utils.render(e)
        b = f"\n{b}\n" if isinstance(e, (Itemize, Enumerate)) else b.strip('"')
        body.append(f"[{b}]")
    rendered = ",\n".join(itertools.chain(args, body)).replace("\n", "\n  ")
    rendered = f"#{function_name}(\n  {rendered}\n)"

    return rendered


def _render_markdown(elements: list[Renderable | str], prefix: str) -> str:
    body = []
    for e in elements:
        b = utils.render(e)
        if isinstance(e, (Itemize, Enumerate)):
            b = b.replace("\n", "\n  ")
            b = "  " + b
        else:
            b = b.strip('"')
            b = f"{prefix} {b}"
        body.append(b)
    rendered = "\n".join(body)

    return rendered



from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any


class Renderable(ABC):
    """
    Abstract base class for any renderable object.

    Any Typst document element should inherit/implement this abstract base class.
    """

    @abstractmethod
    def render(self) -> str:
        pass

    @classmethod
    def __subclasshook__(cls, subclass: Any) -> bool:
        if cls is Renderable:
            if any("render" in B.__dict__ for B in subclass.__mro__):
                return True

        return NotImplemented  # type: ignore[no-any-return]


@dataclass
class Plain:
    """
    Plain string content that is rendered as-is.
    """

    value: str

    def render(self) -> str:
        """
        Render the internal string value as-is.

        Returns:
            The internal string value as-is.
        """
        return self.value



import json
import re
from collections.abc import Callable, Iterable, Mapping, Sequence
from dataclasses import Field, dataclass, fields, is_dataclass
from datetime import date, datetime, timedelta
from typing import Any

from pypst.renderable import Renderable


def render(
    obj: Renderable
    | bool
    | int
    | float
    | str
    | Sequence[Any]
    | Mapping[str, Any]
    | date
    | datetime
    | timedelta
    | None,
) -> str:
    """
    Render renderable objects using their `render` method
    or use the `render_type` utility to render built-in Python types.
    """
    if isinstance(obj, Renderable):
        rendered = obj.render()
    else:
        rendered = render_type(obj)

    return rendered


def render_code(
    obj: Renderable
    | bool
    | int
    | float
    | str
    | Sequence[Any]
    | Mapping[str, Any]
    | date
    | datetime
    | timedelta
    | None,
) -> str:
    """
    Render renderable objects using the `render` method
    and strip any `#` code-mode prefixes.
    """
    return render(obj).lstrip("#")


def render_type(
    arg: bool
    | int
    | float
    | str
    | Sequence[Any]
    | Mapping[str, Any]
    | date
    | datetime
    | timedelta
    | None,
) -> str:
    """
    Render different built-in Python types.
    """
    if arg is None:
        rendered_arg = "none"
    elif isinstance(arg, bool):
        rendered_arg = str(arg).lower()
    elif isinstance(arg, int | float):
        rendered_arg = str(arg)
    elif isinstance(arg, str):
        rendered_arg = arg
    elif isinstance(arg, Sequence):
        rendered_arg = render_sequence(arg)
    elif is_dataclass(arg):
        rendered_arg = render_dataclass(arg)
    elif isinstance(arg, Mapping):
        rendered_arg = render_mapping(arg)
    elif isinstance(arg, date):
        rendered_arg = render_datetime(arg)
    elif isinstance(arg, timedelta):
        rendered_arg = render_timedelta(arg)
    else:
        raise ValueError(f"Invalid argument type: {type(arg)}")

    return rendered_arg


def render_mapping(arg: Mapping[str, Any]) -> str:
    """
    Render a mapping from string to any object supported by `render`.
    """
    rendered = render_sequence(f"{k}: {render_code(v)}" for (k, v) in arg.items())
    if rendered == "()":
        return "(:)"
    return rendered


def render_sequence(arg: Iterable[Any]) -> str:
    """
    Render a sequence of any object supported by `render`.
    """
    return f"({', '.join(render_code(a) for a in arg)})"


def render_datetime(arg: date | datetime) -> str:
    """
    Render a Python `datetime.date` into a call to the Typst datetime function.
    """
    obj = {
        name: getattr(arg, name)
        for name in ["year", "month", "day", "hour", "minute", "second"]
        if hasattr(arg, name)
    }
    return f"#datetime{render_mapping(obj)}"


def render_timedelta(arg: timedelta) -> str:
    """
    Render a Python `datetime.timedelta` into a call to the Typst duration function.
    """
    obj = {
        name: getattr(arg, name)
        for name in [
            "microseconds",
            "seconds",
            "days",
        ]
        if hasattr(arg, name)
    }
    obj["seconds"] = obj["seconds"] + round(obj.pop("microseconds") / 1e6)
    return f"#duration{render_mapping(obj)}"


def render_dataclass(arg: Any) -> str:
    """
    Render a dataclass as either a mapping or a function call.

    The existence and value of an `__is_function__` member on the object
    determines whether it is rendered as a function call. Setting it to return

    Rendering of positional arguments in a Typst function call can be accomplished
    by setting the `positional=True` flag on the dataclass field's metadata.

    You can specify that `None` values should be rendered instead
    of being skipped with `keep_none=True` on the dataclass field's metadata.

    Examples:
        >>> from dataclasses import dataclass, field
        >>> @dataclass
        ... class Foo:
        ...    bar: int | None = 4
        ...    qux: int | None = field(default=16, metadata={"keep_none": True})
        >>>
        >>> render(Foo())
        '#(bar: 4, qux: 16)'
        >>> render(Foo(bar=None))
        '#(qux: 16)'
        >>> render(Foo(qux=None))
        '#(bar: 4, qux: none)'

        >>> @dataclass
        ... class FooFn:
        ...    __is_function__ = True
        ...    bar: int | None = field(metadata={"positional": True})
        ...    qux: int | None = field(default=16, metadata={"keep_none": True})
        >>>
        >>> render(FooFn(4))
        '#foo-fn(4, qux: 16)'
        >>> render(FooFn(None))
        '#foo-fn(qux: 16)'
        >>> render(FooFn(4, qux=None))
        '#foo-fn(4, qux: none)'
    """

    function = getattr(arg, "__is_function__", False)
    fields = dataclass_fields_to_render(arg)

    def name(field: Field) -> str:
        return field.metadata.get("name", field.name)

    def value(field: Field) -> str:
        return render_code(getattr(arg, field.name))

    if function:
        function = (
            function
            if isinstance(function, str)
            else camel_to_kebab_case(arg.__class__.__name__)
        )

        def is_positional(field: Field) -> bool:
            return field.metadata.get("positional", False)

        arguments = ", ".join(
            value(field) if is_positional(field) else f"{name(field)}: {value(field)}"
            for field in fields
        )

        rendered = f"#{function}({arguments})"
    else:
        mapping = render_mapping(
            {name(field): getattr(arg, field.name) for field in fields}
        )
        rendered = f"#{mapping}"

    return rendered


def dataclass_fields_to_render(arg: Any) -> Iterable[Field]:
    """
    These dataclass fields should be rendered.
    Defaults to skipping any fields with a value of `None`.
    """

    def check(field: Field) -> bool:
        # Check whether the field should be skipped.
        if field.metadata.get("skip", False):
            return False

        # When the value is `None`, check whether that should be kept.
        if getattr(arg, field.name) is None:
            return field.metadata.get("keep_none", False)

        # Keep otherwise.
        return True

    return filter(check, fields(arg))


def camel_to_kebab_case(arg: str) -> str:
    """
    Transform a CamelCase string to kebab-case.

    Examples:
        >>> camel_to_kebab_case("ClassName")
        'class-name'
    """
    return re.sub(
        r"([a-z0-9])([A-Z])",
        r"\1-\2",
        arg,
    ).lower()


def render_fenced(
    body: str | Renderable | Iterable[str | Renderable] | None = None,
    context: bool = False,
    start: str = "{",
    end: str = "}",
    indent: int = 2,
    joint: str = "\n",
    render_fn: Callable[[Any], str] = render_code,
) -> str:
    """
    Render a Typst fenced code block such as #{} and #[].

    Example:
        >>> render_fenced()
        '#{}'

        >>> from pypst import Content, SetRule
        >>> string = render_fenced([SetRule("text", {"fill": "red"}), Content("This text will be red.")])
        >>> print(string)
        #{
          set text(fill: red)
          [This text will be red.]
        }
    """

    ctx = "context " if context else ""

    if isinstance(body, (str, Renderable)):
        body = render_fn(body)
    elif isinstance(body, Iterable):
        body = joint.join(render_fn(entry) for entry in body)
    elif body is None:
        body = ""
    else:
        body = render_fn(body)

    if indent is None:
        return f"#{ctx}{start}{body}{end}"

    if "\n" in body:
        newline = f"\n{indent * ' '}"
        indented = body.replace("\n", newline)
        return f"#{ctx}{start}{newline}{indented}\n{end}"

    return f"#{ctx}{start}{body}{end}"


class _RenderDataclass:
    """
    Helper class that implements default dataclass rendering for dictionaries and functions.
    """

    def __str__(self) -> str:
        return self.render()

    def render(self) -> str:
        """
        Render this dataclass as a string. See `render_dataclass` for more information.
        """
        return render_dataclass(self)


class Dictionary(_RenderDataclass):
    """
    Helper class that implements default dataclass rendering for dictionaries in Typst.

    Examples:
        >>> from dataclasses import dataclass, field
        >>> @dataclass
        ... class Foo(Dictionary):
        ...    bar: int | None = 4
        ...    qux: int | None = field(default=16, metadata={"keep_none": True})
        >>>
        >>> Foo().render()
        '#(bar: 4, qux: 16)'
        >>> Foo(bar=None).render()
        '#(qux: 16)'
        >>> Foo(qux=None).render()
        '#(bar: 4, qux: none)'
    """


class Function(_RenderDataclass):
    """
    Helper class that implements default dataclass rendering for function calls in Typst.

    By default, the class name is converted to kebab-case as the function name.
    You can override this behavior by setting the `__is_function__` class variable
    to a string of your choice.

    Examples:
        >>> from dataclasses import dataclass, field
        >>> @dataclass
        ... class FooFn(Function):
        ...    __is_function__ = True
        ...    bar: int | None = field(metadata={"positional": True})
        ...    qux: int | None = field(default=16, metadata={"keep_none": True})
        >>>
        >>> FooFn(4).render()
        '#foo-fn(4, qux: 16)'
        >>> FooFn(None).render()
        '#foo-fn(qux: 16)'
        >>> FooFn(4, qux=None).render()
        '#foo-fn(4, qux: none)'
    """

    __is_function__: bool | str = True


@dataclass
class String:
    """
    String helper that wraps any content in quotes.

    Example:
        >>> line = (
        ...     "The most 'common' string in programming is: "
        ...     '"Hello world!".'
        ... )
        >>> s = String(line)
        >>> print(s.render())
        #"The most 'common' string in programming is: \\"Hello world!\\"."
    """

    body: str | Renderable | None = None

    def __post_init__(self) -> None:
        if isinstance(self.body, String):
            self.body = self.body.body

    def render(self) -> str:
        """
        Render the internal body to a string, escaping any symbols in JSON fashion.
        """
        body = json.dumps(
            "" if self.body is None else render_code(self.body), ensure_ascii=False
        )
        # Always assume we're in content mode.
        return f"#{body}"



from dataclasses import dataclass, field
from typing import Literal

from pypst.document import Document
from pypst.renderable import Renderable
from pypst.utils import Function


@dataclass
class Figure(Function):
    """
    A figure is a block element that contains another element.

    It can be used to add captions, numbering, and other metadata to the element.

    Args:
        body: The element to be rendered inside the figure.
        placement: The placement of the figure.
        caption: The caption of the figure.
        kind: The kind of the figure.
        supplement: The supplement of the figure added when the figure is referenced.
        numbering: The numbering scheme of the figure.
        gap: The gap between body and caption.
        outlined: Whether the figure is added to the outline.

    Examples:
        >>> fig = Figure("[Hello, World!]", caption="[This is content.]")
        >>> print(fig.render())
        #figure([Hello, World!], caption: [This is content.])

        >>> from pypst import Image
        >>> fig = Figure(Image("image.png"), caption="[This is an image.]")
        >>> print(fig.render())
        #figure(image("image.png"), caption: [This is an image.])
    """

    __is_function__ = True

    body: Renderable | str = field(metadata={"positional": True})

    placement: Literal["auto", "top", "bottom"] | None = None
    caption: Renderable | str | None = None
    kind: Renderable | str | None = None
    supplement: Renderable | str | None = None
    numbering: Renderable | str | None = None
    gap: Renderable | str | None = None
    outlined: bool | None = None

    def __post_init__(self) -> None:
        if not isinstance(self.body, (Renderable, str)):
            raise ValueError(f"Invalid body type: {type(self.body)}")
        elif isinstance(self.body, Document):
            raise ValueError(
                "Document cannot be set as body, because it needs to be top-level"
            )



from collections.abc import Mapping
from dataclasses import dataclass
from typing import Any

from pypst.renderable import Renderable
from pypst.utils import render_code


@dataclass
class SetRule:
    """
    Typst set rule.

    Examples:
        >>> rule = SetRule(selector="text", arguments={"fill": "red"})
        >>> rule.render()
        '#set text(fill: red)'

        >>> from pypst import ShowRule
        >>> rule = ShowRule(selector="heading", body=SetRule("text", {"fill": "red"}))
    """

    selector: str | Renderable
    arguments: Mapping[str, Any] | Renderable

    def __str__(self) -> str:
        return self.render()

    def render(self) -> str:
        """
        Render the set rule to a string.
        """
        return f"#set {render_code(self.selector)}{render_code(self.arguments)}"



from typing import Any

from pypst.cell import Cell
from pypst.content import Content
from pypst.document import Document
from pypst.figure import Figure
from pypst.functional import Functional
from pypst.heading import Heading
from pypst.image import Image
from pypst.itemize import Enumerate, Itemize
from pypst.renderable import Plain, Renderable
from pypst.set_rule import SetRule
from pypst.show_rule import ShowRule

try:
    from pypst.table import Table  # needed to be included in code completion
except ModuleNotFoundError:
    pass


def __getattr__(name: str) -> Any:
    """Lazily import Table to check for pandas."""
    if name == "Table":
        from pypst.table import Table

        return Table
    raise AttributeError(f"module {__name__!r} has no attribute {name!r}")


__all__ = [
    "Cell",
    "Content",
    "Document",
    "Enumerate",
    "Figure",
    "Functional",
    "Heading",
    "Image",
    "Itemize",
    "Plain",
    "Renderable",
    "SetRule",
    "ShowRule",
    "Table",
]
__version__ = "0.7.0"



import itertools
from dataclasses import dataclass
from typing import Optional, Union, Literal

try:
    import pandas as pd
except ImportError:
    raise ModuleNotFoundError(
        "Could not find Pandas. To use the table functionality, Pypst needs to be installed as 'pypst[pandas]'."
    )
from frozendict import frozendict

from frozenlist import FrozenList

from pypst.cell import Cell
from pypst.utils import render_type, render_mapping


@dataclass
class Table:
    """
    A table element.

    A table is constructed from a pandas DataFrame using the [from_dataframe][pypst.table.Table.from_dataframe] method.
    Afterward, styling options can be set using the properties of the class.
    Manual horizontal and vertical lines can be added using the
    `add_hline` and `add_vline` methods.

    Most properties that accept a string also accept functions in Typst.
    Typst functions can be supplied in string form,
    for example, `"(x, y) => if x > y then {red} else {black}"`.

    Hint:
        The header, index and row data can be accessed via the `header_data`,
        `index_data` and `row_data` properties.
        The lists can't be modified directly,
        but the elements can be accessed to apply styling to individual cells.
    """

    header_data: FrozenList[FrozenList["Cell"]] = FrozenList(FrozenList([]))
    index_data: FrozenList[FrozenList["Cell"]] = FrozenList(FrozenList([]))
    row_data: FrozenList[FrozenList["Cell"]] = FrozenList(FrozenList([]))
    _columns: Optional[int | str | FrozenList[str]] = None
    _rows: Optional[int | str | FrozenList[str]] = None
    _stroke: Optional[str | FrozenList[str] | frozendict[str, str]] = None
    _align: Optional[str | FrozenList[str]] = None
    _fill: Optional[str | FrozenList[str]] = None
    _gutter: Optional[int | str | FrozenList[str]] = None
    _column_gutter: Optional[int | str | FrozenList[str]] = None
    _row_gutter: Optional[int | str | FrozenList[str]] = None
    _lines: Optional[list["TableLine"]] = None

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame) -> "Table":
        """
        Create a new table element from a Pandas DataFrame.

        The `columns` property of the data frame will be used as the header data and
        the `index` property will be used as the index data.
        The header is rendered as the `table.header` element and the index as the
        first columns in each row.
        Multi-level indexes are supported.

        After creating the table, styling options can be set using the properties of
        the table class.

        Args:
            df: The DataFrame to create the table from.

        Examples:
            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> print(table.render())
            #table(
              columns: 3,
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )

            >>> df = pd.DataFrame(
            ...     [[1, 3], [2, 4]],
            ...     columns=pd.MultiIndex.from_tuples([("A", "mean"), ("A", "std")]),
            ...     index=["X", "Y"]
            ... )
            >>> table = Table.from_dataframe(df)
            >>> print(table.render())
            #table(
              columns: 3,
              table.header[#table.cell(rowspan: 2)[]][#table.cell(colspan: 2)[A]][mean][std],
              [X], [1], [3],
              [Y], [2], [4]
            )

            >>> df = pd.DataFrame(
            ...     [[1, 3], [2, 4]],
            ...     columns=["A", "B"],
            ...     index=pd.MultiIndex.from_tuples([("X", "mean"), ("X", "std")])
            ... )
            >>> table = Table.from_dataframe(df)
            >>> print(table.render())
            #table(
              columns: 4,
              table.header[#table.cell(colspan: 2)[]][A][B],
              [#table.cell(rowspan: 2)[X]], [mean], [1], [3],
              [std], [2], [4]
            )

        Returns:
            The new table element.
        """
        table = cls()
        table.header_data = _parse_index(df.columns, direction="cols")
        table.index_data = _parse_index(df.index, direction="rows")
        row_data: FrozenList[FrozenList[Cell]] = FrozenList([])
        for _, *row in df.itertuples():
            row_cells = FrozenList([Cell(value) for value in row])
            row_cells.freeze()
            row_data.append(row_cells)
        row_data.freeze()
        table.row_data = row_data
        table.columns = len(df.columns) + df.index.nlevels

        return table

    @property
    def columns(self) -> Optional[int | str | FrozenList[str]]:
        """
        The number of columns or column styling array.

        This property is automatically set when creating a table from a DataFrame.
        It can be replaced by an integer to apply styling to all columns or by a list
        to apply styling to individual columns.

        If a list is provided, it must have the same length as the number of columns
        plus the number of index levels.

        Examples:
            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.columns = "50%"
            >>> print(table.render())
            #table(
              columns: 50%,
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )

            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.columns = ["10pt", "20pt", "30pt"] # two columns plus index
            >>> print(table.render())
            #table(
              columns: (10pt, 20pt, 30pt),
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )
        """
        return self._columns

    @columns.setter
    def columns(self, value: int | str | list[str]) -> None:
        if not isinstance(value, (int, str, list)):
            raise ValueError("Columns must be an integer, string, or list of strings")
        elif isinstance(value, list) and not all(isinstance(v, str) for v in value):
            raise ValueError("All elements in the list must be strings")
        elif isinstance(value, int) and not value == (
            len(self.row_data[0]) + len(self.index_data)
        ):
            raise ValueError(
                "Number of columns must match the number "
                "of table columns plus index levels"
            )
        elif isinstance(value, list) and not len(value) == (
            len(self.row_data[0]) + len(self.index_data)
        ):
            raise ValueError(
                "If specifying columns as a list, "
                "its length must match the number of table columns plus index levels"
            )
        if isinstance(value, list):
            self._columns = FrozenList(value)
            self._columns.freeze()
        else:
            self._columns = value

    @property
    def rows(self) -> Optional[int | str | FrozenList[str]]:
        """
        The number of rows or row styling array.

        This property is automatically set when creating a table from a DataFrame.
        It can be replaced by an integer to apply styling to all rows or by a list
        to apply styling to individual rows.

        If a list is provided, it must have the same length as the number of rows
        plus the number of header levels.

        Examples:
            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.rows = "50%"
            >>> print(table.render())
            #table(
              columns: 3,
              rows: 50%,
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )

            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.rows = ["10pt", "20pt", "30pt"] # two rows plus header
            >>> print(table.render())
            #table(
              columns: 3,
              rows: (10pt, 20pt, 30pt),
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )
        """
        return self._rows

    @rows.setter
    def rows(self, value: int | str | list[str]) -> None:
        if not isinstance(value, (int, str, list)) and value is not None:
            raise ValueError(
                "Rows must be an integer, string, list of strings, or None"
            )
        elif isinstance(value, list) and not all(isinstance(v, str) for v in value):
            raise ValueError("All elements in the list must be strings")
        elif isinstance(value, int) and not value == (
            len(self.row_data) + len(self.header_data)
        ):
            raise ValueError(
                "Number of rows must match the number of table rows plus header levels"
            )
        elif isinstance(value, list) and not len(value) == (
            len(self.row_data) + len(self.header_data)
        ):
            raise ValueError(
                "If specifying rows as a list, "
                "its length must match the number of table rows plus header levels"
            )

        if isinstance(value, list):
            self._rows = FrozenList(value)
            self._rows.freeze()
        else:
            self._rows = value

    @property
    def stroke(self) -> Optional[str | FrozenList[str] | frozendict[str, str]]:
        """
        The stroke style for the table cells.

        This can be either a string to style all cells,
        a list of strings to style each column,
        or a dictionary to control the stroke of each of the cells' sides.

        Examples:
            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.stroke = "2pt"
            >>> print(table.render())
            #table(
              columns: 3,
              stroke: 2pt,
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )

            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.stroke = ["2pt", "1pt", "3pt"]
            >>> print(table.render())
            #table(
              columns: 3,
              stroke: (2pt, 1pt, 3pt),
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )

            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.stroke = {"top": "2pt", "bottom": "1pt"}
            >>> print(table.render())
            #table(
              columns: 3,
              stroke: (top: 2pt, bottom: 1pt),
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )
        """
        return self._stroke

    @stroke.setter
    def stroke(self, value: Optional[str | FrozenList[str] | dict[str, str]]) -> None:
        if not isinstance(value, (str, list, dict)) and value is not None:
            raise ValueError(
                "Stroke must be a string, list of strings, dictionary, or None"
            )
        elif isinstance(value, list) and not all(isinstance(v, str) for v in value):
            raise ValueError("All elements in the list must be strings")
        elif isinstance(value, dict) and not all(
            isinstance(k, str) and isinstance(v, str) for k, v in value.items()
        ):
            raise ValueError("All keys and values in the dictionary must be strings")

        if isinstance(value, list):
            self._stroke = FrozenList(value)
            self._stroke.freeze()
        elif isinstance(value, dict):
            self._stroke = frozendict(value)
        else:
            self._stroke = value

    @property
    def align(self) -> Optional[str | FrozenList[str]]:
        """
        The alignment for the table columns.

        This can be either a string to align all columns,
        or a list of strings to align individual columns.

        Examples:
            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.align = "right"
            >>> print(table.render())
            #table(
              columns: 3,
              align: right,
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )

            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.align = ["left", "center", "right"]  # two columns plus index
            >>> print(table.render())
            #table(
              columns: 3,
              align: (left, center, right),
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )
        """
        return self._align

    @align.setter
    def align(self, value: Optional[str | list[str]]) -> None:
        if not isinstance(value, (str, list)) and value is not None:
            raise ValueError("Align must be a string, list of strings, or None")
        elif isinstance(value, list) and not all(isinstance(v, str) for v in value):
            raise ValueError("All elements in the align list must be strings")
        if isinstance(value, list):
            self._align = FrozenList(value)
            self._align.freeze()
        else:
            self._align = value

    @property
    def fill(self) -> Optional[str | FrozenList[str]]:
        """
        The background color for the table cells.

        This can be either a string to fill all cells,
        or a list of strings to fill individual columns.

        Examples:
            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.fill = "red"
            >>> print(table.render())
            #table(
              columns: 3,
              fill: red,
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )

            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.fill = ["red", "blue", "green"]  # two columns plus index
            >>> print(table.render())
            #table(
              columns: 3,
              fill: (red, blue, green),
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )
        """
        return self._fill

    @fill.setter
    def fill(self, value: Optional[str | list[str]]) -> None:
        if not isinstance(value, (str, list)) and value is not None:
            raise ValueError("Fill must be a string, list of strings, or None")
        elif isinstance(value, list) and not all(isinstance(v, str) for v in value):
            raise ValueError("All elements in the fill list must be strings")
        if isinstance(value, list):
            self._fill = FrozenList(value)
            self._fill.freeze()
        else:
            self._fill = value

    @property
    def gutter(self) -> Optional[int | str | FrozenList[str]]:
        """
        The spacing between table cells.

        This can be either an integer/string to set the gutter for all cells,
        or a list of strings to set the gutter for individual columns.

        Examples:
            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.gutter = 10
            >>> print(table.render())
            #table(
              columns: 3,
              gutter: 10,
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )

            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.gutter = ["10pt", "20pt", "30pt"]  # two columns plus index
            >>> print(table.render())
            #table(
              columns: 3,
              gutter: (10pt, 20pt, 30pt),
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )
        """
        return self._gutter

    @gutter.setter
    def gutter(self, value: Optional[int | str | list[str]]) -> None:
        if not isinstance(value, (int | str, list)) and value is not None:
            raise ValueError("Gutter must be an int, string, list of strings, or None")
        elif isinstance(value, list) and not all(isinstance(v, str) for v in value):
            raise ValueError("All elements in the gutter list must be strings")
        if isinstance(value, list):
            self._gutter = FrozenList(value)
            self._gutter.freeze()
        else:
            self._gutter = value

    @property
    def column_gutter(self) -> Optional[int | str | FrozenList[str]]:
        """
        The spacing between table columns.

        This can be either an integer/string to set the column gutter for all columns,
        or a list of strings to set the column gutter for individual columns.

        Examples:
            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.column_gutter = 10
            >>> print(table.render())
            #table(
              columns: 3,
              column-gutter: 10,
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )

            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.column_gutter = ["10pt", "20pt", "30pt"]  # two columns plus index
            >>> print(table.render())
            #table(
              columns: 3,
              column-gutter: (10pt, 20pt, 30pt),
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )
        """
        return self._column_gutter

    @column_gutter.setter
    def column_gutter(self, value: Optional[int | str | list[str]]) -> None:
        if not isinstance(value, (int | str, list)) and value is not None:
            raise ValueError(
                "Column gutter must be an int, string, list of strings, or None"
            )
        elif isinstance(value, list) and not all(isinstance(v, str) for v in value):
            raise ValueError("All elements in the column gutter list must be strings")
        if isinstance(value, list):
            self._column_gutter = FrozenList(value)
            self._column_gutter.freeze()
        else:
            self._column_gutter = value

    @property
    def row_gutter(self) -> Optional[int | str | FrozenList[str]]:
        """
        The spacing between table rows.

        This can be either an integer/string to set the row gutter for all cells,
        or a list of strings to set the row gutter for individual columns.

        Examples:
            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.row_gutter = 10
            >>> print(table.render())
            #table(
              columns: 3,
              row-gutter: 10,
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )

            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.row_gutter = ["10pt", "20pt", "30pt"]  # two columns plus index
            >>> print(table.render())
            #table(
              columns: 3,
              row-gutter: (10pt, 20pt, 30pt),
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )
        """
        return self._row_gutter

    @row_gutter.setter
    def row_gutter(self, value: Optional[int | str | list[str]]) -> None:
        if not isinstance(value, (int | str, list)) and value is not None:
            raise ValueError(
                "Row gutter must be an int, string, list of strings, or None"
            )
        elif isinstance(value, list) and not all(isinstance(v, str) for v in value):
            raise ValueError("All elements in the row gutter list must be strings")
        if isinstance(value, list):
            self._row_gutter = FrozenList(value)
            self._row_gutter.freeze()
        else:
            self._row_gutter = value

    def add_hline(
        self,
        y: int,
        start: Optional[int] = None,
        end: Optional[int] = None,
        stroke: Optional[str] = None,
        position: Optional[Literal["start", "end"]] = None,
    ) -> None:
        """
        Add a horizontal line to the table at the specified row position.

        Args:
            y: The row position.
            start: The column to start the line.
            end: The column to end the line.
            stroke: The stroke style for the line.
            position: Whether the line should appear before or after the row.

        Examples:
            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.add_hline(1, stroke="2pt")
            >>> print(table.render())
            #table(
              columns: 3,
              table.hline(y: 1, stroke: 2pt),
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )
        """
        self._add_line(TableLine(y, "horizontal", start, end, stroke, position))

    def add_vline(
        self,
        x: int,
        start: Optional[int] = None,
        end: Optional[int] = None,
        stroke: Optional[str] = None,
        position: Optional[Literal["start", "end"]] = None,
    ) -> None:
        """
        Add a vertical line to the table at the specified column position.

        Args:
            x: The column position.
            start: The row to start the line.
            end: The row to end the line.
            stroke: The stroke style for the line.
            position: Whether the line should appear before or after the column.

        Examples:
            >>> df = pd.DataFrame({"A": [1, 2], "B": [3, 4]}, index=["X", "Y"])
            >>> table = Table.from_dataframe(df)
            >>> table.add_vline(1, stroke="2pt")
            >>> print(table.render())
            #table(
              columns: 3,
              table.vline(x: 1, stroke: 2pt),
              table.header[][A][B],
              [X], [1], [3],
              [Y], [2], [4]
            )
        """
        self._add_line(TableLine(x, "vertical", start, end, stroke, position))

    def _add_line(self, line: "TableLine") -> None:
        if self._lines is None:
            self._lines = []
        self._lines.append(line)

    def __str__(self) -> str:
        return self.render()

    def __repr__(self) -> str:
        return f"Table(headers={self.header_data}, index={self.index_data}, rows={self.row_data})"

    def render(self) -> str:
        """
        Render the table element to a string.

        Vertical and horizontal lines are rendered before the table body.
        All attributes and the body are indented by two spaces.

        Returns:
            The rendered table element.
        """
        headers = itertools.chain(*self.header_data)
        index_placeholder = Cell(
            rowspan=len(self.header_data), colspan=len(self.index_data)
        )
        inner = (
            self._render_args()
            + self._render_lines()
            + "table.header"
            + index_placeholder.render()
            + "".join(header.render() for header in headers)
            + ",\n"
            + self._render_rows()
        )
        # indent body and args by 2 spaces
        table = "#table(\n  " + inner.replace("\n", "\n  ") + "\n)"

        return table

    def _render_args(self) -> str:
        args = []
        if self._columns is not None:
            columns = render_type(self._columns)
            args.append(f"columns: {columns}")
        if self._rows is not None:
            rows = render_type(self._rows)
            args.append(f"rows: {rows}")
        if self._stroke is not None:
            stroke = render_type(self._stroke)
            args.append(f"stroke: {stroke}")
        if self._align is not None:
            align = render_type(self._align)
            args.append(f"align: {align}")
        if self._fill is not None:
            fill = render_type(self._fill)
            args.append(f"fill: {fill}")
        if self._gutter is not None:
            gutter = render_type(self._gutter)
            args.append(f"gutter: {gutter}")
        if self._column_gutter is not None:
            column_gutter = render_type(self._column_gutter)
            args.append(f"column-gutter: {column_gutter}")
        if self._row_gutter is not None:
            row_gutter = render_type(self._row_gutter)
            args.append(f"row-gutter: {row_gutter}")
        rendered_args = ",\n".join(args) + ",\n"

        return rendered_args

    def _render_lines(self) -> str:
        if not self._lines:
            rendered_lines = ""
        else:
            rendered_lines = ",\n".join(line.render() for line in self._lines) + ",\n"

        return rendered_lines

    def _render_rows(self) -> str:
        table = ""
        rows_to_skip = [0] * len(self.index_data)
        index_positions = [0] * len(self.index_data)
        for row in self.row_data:
            for level in range(len(self.index_data)):
                if rows_to_skip[level] == 0:
                    index_cell = self.index_data[level][index_positions[level]]
                    table += index_cell.render() + ", "
                    index_positions[level] += 1
                    rows_to_skip[level] = index_cell.rowspan
                rows_to_skip[level] -= 1
            table += ", ".join(cell.render() for cell in row) + ",\n"
        table = table[:-2]  # remove trailing comma and newline

        return table


@dataclass
class TableLine:
    pos: int
    orientation: Literal["horizontal", "vertical"]
    start: Optional[int] = None
    end: Optional[int] = None
    stroke: Optional[str | dict[str, str]] = None
    position: Optional[Literal["start", "end"]] = None

    def render(self) -> str:
        args = []
        if self.orientation == "horizontal":
            line = "table.hline({0})"
            args.append(f"y: {self.pos}")
        else:
            line = "table.vline({0})"
            args.append(f"x: {self.pos}")

        if self.start is not None:
            args.append(f"start: {self.start}")
        if self.end is not None:
            args.append(f"end: {self.end}")
        if self.stroke is not None:
            stroke = self.stroke
            if isinstance(self.stroke, dict):
                stroke = render_mapping(self.stroke)
            args.append(f"stroke: {stroke}")
        if self.position is not None:
            args.append(f"position: {self.position}")
        line = line.format(", ".join(args))

        return line


def _parse_index(
    index: Union[pd.Index, pd.MultiIndex], direction: Literal["rows", "cols"]
) -> FrozenList[FrozenList["Cell"]]:
    headers: list[list[Cell]]
    if index.empty:
        headers = [[]]
    elif isinstance(index, pd.MultiIndex):
        headers = _parse_multi_index(index, direction)
    else:
        headers = [[Cell(header, colspan=1) for header in index]]

    frozen_headers = FrozenList(FrozenList(header) for header in headers)
    for header in frozen_headers:
        header.freeze()
    frozen_headers.freeze()

    return frozen_headers


def _parse_multi_index(
    index: pd.MultiIndex, direction: Literal["rows", "cols"]
) -> list[list["Cell"]]:
    assert not index.empty
    headers = []
    for level in range(index.nlevels):
        headers.append(_parse_level(index, level, direction))

    return headers


def _parse_level(
    index: pd.MultiIndex, level: int, direction: Literal["rows", "cols"]
) -> list["Cell"]:
    span_arg = _get_span_arg(direction)
    level_codes = index.codes[level]
    headers = []
    span = 0
    prev_code = level_codes[0]
    for code in level_codes:
        if code == prev_code:
            span += 1
        else:
            headers.append(Cell(index.levels[level][prev_code]))
            setattr(headers[-1], span_arg, span)
            prev_code = code
            span = 1
    headers.append(Cell(index.levels[level][-1]))
    setattr(headers[-1], span_arg, span)

    return headers


def _get_span_arg(direction: Literal["rows", "cols"]) -> str:
    if direction == "rows":
        span_arg = "rowspan"
    elif direction == "cols":
        span_arg = "colspan"
    else:
        raise ValueError(f"Invalid direction: {direction}")

    return span_arg



from dataclasses import dataclass

from pypst.content import Content
from pypst.functional import Functional
from pypst.renderable import Renderable
from pypst.utils import render_code


@dataclass
class ShowRule:
    """
    Typst show rule.

    Examples:
        >>> rule = ShowRule(selector="heading", body="text.with(fill: red)")
        >>> rule.render()
        '#show heading: text.with(fill: red)'

        >>> rule = ShowRule(selector="heading", argument="it", body="text(fill: red, it)")
        >>> print(rule.render())
        #show heading: it => text(fill: red, it)

        >>> rule = ShowRule(selector="heading", argument="it", body=Functional(["v(0.5em)", "it"]))
        >>> print(rule.render())
        #show heading: it => {
          v(0.5em)
          it
        }

        >>> rule = ShowRule(selector="heading", argument="it", body=Content(["#v(0.5em)", "#it"]))
        >>> print(rule.render())
        #show heading: it => [
          #v(0.5em)
          #it
        ]
    """

    selector: str | Renderable | None = None
    argument: str | Renderable | None = None
    body: str | Renderable | None = None
    indent: int = 2

    def __post_init__(self) -> None:
        if not isinstance(self.body, (Functional, Content)):
            body = Functional(self.body, indent=self.indent)
            assert isinstance(body, Renderable)
            self.body = body

    def __str__(self) -> str:
        return self.render()

    def render(self) -> str:
        """
        Render the show rule to a string.
        """
        target = "" if self.selector is None else f" {self.selector}"

        if self.body is None:
            body = "none"
        else:
            # Render functional or content block.
            body = render_code(self.body)

            # Remove brackets from one-liner functional blocks that don't initiate a context.
            if (
                isinstance(self.body, Functional)
                and not self.body.context
                and "\n" not in body
            ):
                body = body[1:-1]

        if self.argument is None:
            return f"#show{target}: {body}"
        else:
            argument = render_code(self.argument)
            return f"#show{target}: {argument} => {body}"



from collections.abc import Iterable
from dataclasses import dataclass

from pypst.renderable import Renderable
from pypst.utils import render, render_fenced


@dataclass
class Content:
    """
    Typst content block explicitly wrapped in [].

    Example:
        >>> content = Content("Foo is a bar.")
        >>> content.render()
        '#[Foo is a bar.]'

        >>> from pypst import SetRule
        >>> content = Content([SetRule("text", {"fill": "red"}), "This text will be red."])
        >>> print(content.render())
        #[
          #set text(fill: red)
          This text will be red.
        ]

        >>> from pypst import SetRule
        >>> content = Content([SetRule("text", {"fill": "red"}), "This is page #here().page()"], context=True)
        >>> print(content.render())
        #context [
          #set text(fill: red)
          This is page #here().page()
        ]
    """

    body: str | Renderable | Iterable[str | Renderable] | None = None
    context: bool = False
    joint: str = "\n"
    indent: int = 2

    def __post_init__(self) -> None:
        if isinstance(self.body, Content):
            self.context = self.context or self.body.context
            self.body = self.body.body

    def __str__(self) -> str:
        return self.render()

    def render(self) -> str:
        """
        Render the content block to a string.
        """
        return render_fenced(
            body=self.body,
            context=self.context,
            joint=self.joint,
            indent=self.indent,
            start="[",
            end="]",
            render_fn=render,
        )
